import json

from falcon import HTTP_OK, HTTP_METHOD_NOT_ALLOWED, HTTP_BAD_REQUEST, HTTP_NOT_FOUND
from falcon.testing import TestCase

from orchestrator.api import stacks

TEST_INSTANCE_ARGS = {
    'cpu_count': 2,
    'memory': 256 * 1000 * 1000
}

class StackTest(TestCase):
    def setUp(self):
        super(StackTest, self).setUp()
        self.app = stacks.create_app()

    def test_stacks_api_allowed_methods_on_endpoints(self):
        self.assertEqual(HTTP_OK, self.simulate_get('/stacks').status)
        self.assertEqual(HTTP_NOT_FOUND, self.simulate_put('/stacks/42').status)
        self.assertEqual(HTTP_NOT_FOUND, self.simulate_delete('/stacks/42').status)
        self.assertEqual(HTTP_NOT_FOUND, self.simulate_get('/stacks/42').status)
        self.assertEqual(HTTP_BAD_REQUEST, self.simulate_post('/stacks').status)

        self.assertEqual(HTTP_METHOD_NOT_ALLOWED, self.simulate_post('/stacks/42').status)
        self.assertEqual(HTTP_METHOD_NOT_ALLOWED, self.simulate_put('/stacks').status)
        self.assertEqual(HTTP_METHOD_NOT_ALLOWED, self.simulate_delete('/stacks').status)

    def test_stacks_api_create(self):
        # given
        stack_request_body = {
            'id': 'root',
            'type': 'group',
            'resources': [
                {
                    'id': 'instance-1',
                    'type': 'instance',
                    'cpu_count': 1,
                    'memory': 100
                },
                {
                    'id': 'level-1',
                    'type': 'group',
                    'resources': [
                        {
                            'id': 'instance-2',
                            'type': 'instance',
                            'cpu_count': 2,
                            'memory': 200
                        },
                        {
                            'id': 'instance-3',
                            'type': 'instance',
                            'cpu_count': 3,
                            'memory': 300
                        },
                        {
                            'id': 'connection-1',
                            'type': 'connection',
                            'instance_id': 'instance-1',
                            'vlan': 42,
                            'mac': 'aa:bb:cc:dd:ee:ff',
                            'ip_address': '10.0.0.42'
                        }
                    ]
                }
            ]
        }

        # when
        result = self.simulate_post('/stacks', body=json.dumps(stack_request_body))

        # then
        self.assertEqual(sort_resources(stack_request_body), sort_resources(result.json))

    def test_stacks_api_create_with_empty_json(self):
        # given
        empty_request_body = '{}'

        # when
        result = self.simulate_post('/stacks', body=empty_request_body, headers={'content-type': 'application/json'})

        # then
        self.assertIn('id', result.json.keys())
        self.assertIn('type', result.json.keys())
        self.assertIn('resources', result.json.keys())
        self.assertEqual('group', result.json['type'])
        self.assertEqual([], result.json['resources'])

    def test_stacks_api_create_with_empty_body(self):
        # when
        result = self.simulate_post('/stacks', headers={'content-type': 'application/json'})

        # then
        self.assertEqual(HTTP_BAD_REQUEST, result.status)

    def test_stacks_api_create_invalid_input(self):
        # given
        invalid_request_bodies = [
            json.dumps({'invalid': 'value'}),
            json.dumps({'resources': [{'invalid': 'value'}]})
        ]

        for body in invalid_request_bodies:
            # when
            result = self.simulate_post('/stacks', body=body, headers={'content-type': 'application/json'})

            # then
            self.assertEqual(HTTP_BAD_REQUEST, result.status)

    def test_stacks_api_update(self):
        # given
        stack_id = 'some_resource_group'
        resource_id = 'new_resource'
        group1 = {'id': stack_id, 'type': 'group', 'resources': []}
        resource = {'id': resource_id, 'type': 'instance'}
        resource.update(TEST_INSTANCE_ARGS)
        group2 = {'id': stack_id, 'type': 'group', 'resources': [resource]}
        expected_diff = {
            'added': {
                '%s.%s' % (stack_id, resource_id): resource
            },
            'removed': {}
        }

        # when
        self.simulate_post('/stacks', body=json.dumps(group1))
        put_result = self.simulate_put('/stacks/%s' % stack_id, body=json.dumps(group2))
        get_result = self.simulate_get('/stacks')

        # then
        self.assertEqual(expected_diff, put_result.json)
        self.assertIn(group2, get_result.json)

    def test_stacks_api_delete(self):
        # given
        resource_id = 'some_resource'
        resource = {'id': resource_id, 'type': 'instance'}
        resource.update(TEST_INSTANCE_ARGS)

        # when
        self.simulate_post('/stacks', body=json.dumps(resource))
        list_result_before_delete = self.simulate_get('/stacks')
        delete_result = self.simulate_delete('/stacks/%s' % resource_id)
        list_result_after_delete = self.simulate_get('/stacks')

        # then
        self.assertEqual([resource], list_result_before_delete.json)
        self.assertEqual(HTTP_OK, delete_result.status)
        self.assertEqual([], list_result_after_delete.json)

    def test_stacks_api_get_list(self):
        # given
        resource1 = {'id': 'some_resource', 'type': 'instance'}
        resource1.update(TEST_INSTANCE_ARGS)
        resource2 = {'id': 'some_other_resource', 'type': 'instance'}
        resource2.update(TEST_INSTANCE_ARGS)

        # when
        self.simulate_post('/stacks', body=json.dumps(resource1))
        self.simulate_post('/stacks', body=json.dumps(resource2))
        result = self.simulate_get('/stacks').json

        # then
        self.assertEqual(2, len(result))
        self.assertIn(resource1, result)
        self.assertIn(resource2, result)

    def test_stacks_api_get_one(self):
        # given
        resource_id = 'some_resource'
        resource = {'id': resource_id, 'type': 'instance'}
        resource.update(TEST_INSTANCE_ARGS)

        # when
        self.simulate_post('/stacks', body=json.dumps(resource))
        result = self.simulate_get('/stacks/%s' % resource_id).json

        # then
        self.assertEqual([resource], result)


def sort_resources(group):
    try:
        for resource in group['resources']:
            sort_resources(resource)
        group['resources'].sort(key=lambda r: r['id'])
    except (KeyError, AttributeError):
        pass
